@@ -7,8 +7,8 @@
 use super::*;
 use crate::{
     common::{
+        writer::{write_bytes, write_param, write_param_value, write_params, write_value},
         ArchivedPartialDateTime,
-        writer::{write_bytes, write_param_value, write_text},
     },
     icalendar::ValueSeparator,
 };
@@ -20,7 +20,7 @@ use std::{
 impl ArchivedICalendar {
     pub fn write_to(&self, out: &mut impl Write) -> std::fmt::Result {
         let _v = [0.into()];
-        let mut component_iter: Iter<'_, rkyv::rend::u32_le> = _v.iter();
+        let mut component_iter: Iter<'_, rkyv::rend::u16_le> = _v.iter();
         let mut component_stack = Vec::with_capacity(4);
 
         loop {
@@ -74,101 +74,150 @@ impl ArchivedICalendarEntry {
             line_len += 18;
         }
 
-        let mut types = None;
-        let mut last_param: Option<&ArchivedICalendarParameterName> = None;
-
+        let mut value_type = None;
         for param in self.params.iter() {
-            if last_param.is_some_and(|last_param| last_param == &param.name) {
-                write!(out, ",")?;
-                line_len += 1;
+            write!(out, ";")?;
+            line_len += 1;
 
-                if line_len + 1 > 75 {
-                    write!(out, "\r\n ")?;
-                    line_len = 1;
-                }
-            } else {
-                write!(out, ";")?;
-                line_len += 1;
-                let name = param.name.as_str();
-                let need_len = name.len() + 1;
-                if line_len + need_len > 75 {
-                    write!(out, "\r\n ")?;
-                    line_len = 1;
-                }
-                if !matches!(param.value, ArchivedICalendarParameterValue::Null) {
-                    write!(out, "{name}=")?;
-                } else {
-                    write!(out, "{name}")?;
-                }
-                line_len += need_len;
-                last_param = Some(&param.name);
+            if line_len + 1 > 75 {
+                write!(out, "\r\n ")?;
+                line_len = 1;
             }
 
-            match &param.value {
-                ArchivedICalendarParameterValue::Text(v) => {
-                    write_param_value(out, &mut line_len, v)?;
+            match param {
+                ArchivedICalendarParameter::Altrep(v) => {
+                    write_uri_param(out, &mut line_len, "ALTREP", v)?;
                 }
-                ArchivedICalendarParameterValue::Integer(i) => {
-                    write!(out, "{i}")?;
-                    line_len += 2;
+                ArchivedICalendarParameter::Cn(v) => {
+                    write_param(out, &mut line_len, "CN", v)?;
                 }
-                ArchivedICalendarParameterValue::Bool(v) => {
-                    let v = if !matches!(param.name, ArchivedICalendarParameterName::Range) {
-                        if *v { "TRUE" } else { "FALSE" }
-                    } else {
-                        "THISANDFUTURE"
-                    };
-                    line_len += v.len();
-                    write!(out, "{v}")?;
+                ArchivedICalendarParameter::Cutype(v) => {
+                    write_param(out, &mut line_len, "CUTYPE", v)?;
                 }
-                ArchivedICalendarParameterValue::Uri(uri) => {
-                    write!(out, "\"")?;
-                    write_uri(out, &mut line_len, uri, false)?;
-                    write!(out, "\"")?;
+                ArchivedICalendarParameter::DelegatedFrom(v) => {
+                    write_uri_params(out, &mut line_len, "DELEGATED-FROM", v)?;
                 }
-                ArchivedICalendarParameterValue::Cutype(v) => {
-                    write_param_value(out, &mut line_len, v.as_str())?;
+                ArchivedICalendarParameter::DelegatedTo(v) => {
+                    write_uri_params(out, &mut line_len, "DELEGATED-TO", v)?;
                 }
-                ArchivedICalendarParameterValue::Fbtype(v) => {
-                    write_param_value(out, &mut line_len, v.as_str())?;
+                ArchivedICalendarParameter::Dir(v) => {
+                    write_uri_param(out, &mut line_len, "DIR", v)?;
                 }
-                ArchivedICalendarParameterValue::Partstat(v) => {
-                    write_param_value(out, &mut line_len, v.as_str())?;
+                ArchivedICalendarParameter::Fmttype(v) => {
+                    write_param(out, &mut line_len, "FMTTYPE", v)?;
                 }
-                ArchivedICalendarParameterValue::Related(v) => {
-                    write_param_value(out, &mut line_len, v.as_str())?;
+                ArchivedICalendarParameter::Fbtype(v) => {
+                    write_param(out, &mut line_len, "FBTYPE", v)?;
                 }
-                ArchivedICalendarParameterValue::Reltype(v) => {
-                    write_param_value(out, &mut line_len, v.as_str())?;
+                ArchivedICalendarParameter::Language(v) => {
+                    write_param(out, &mut line_len, "LANGUAGE", v)?;
                 }
-                ArchivedICalendarParameterValue::Role(v) => {
-                    write_param_value(out, &mut line_len, v.as_str())?;
+                ArchivedICalendarParameter::Member(v) => {
+                    write_uri_params(out, &mut line_len, "MEMBER", v)?;
                 }
-                ArchivedICalendarParameterValue::ScheduleAgent(v) => {
-                    write_param_value(out, &mut line_len, v.as_str())?;
+                ArchivedICalendarParameter::Partstat(v) => {
+                    write_param(out, &mut line_len, "PARTSTAT", v)?;
                 }
-                ArchivedICalendarParameterValue::ScheduleForceSend(v) => {
-                    write_param_value(out, &mut line_len, v.as_str())?;
+                ArchivedICalendarParameter::Range => {
+                    write!(out, "RANGE=THISANDFUTURE")?;
+                    line_len += 18;
                 }
-                ArchivedICalendarParameterValue::Value(v) => {
-                    types = Some(v);
-                    write_param_value(out, &mut line_len, v.as_str())?;
+                ArchivedICalendarParameter::Related(v) => {
+                    write_param(out, &mut line_len, "RELATED", v)?;
                 }
-                ArchivedICalendarParameterValue::Display(v) => {
-                    write_param_value(out, &mut line_len, v.as_str())?;
+                ArchivedICalendarParameter::Reltype(v) => {
+                    write_param(out, &mut line_len, "RELTYPE", v)?;
                 }
-                ArchivedICalendarParameterValue::Feature(v) => {
-                    write_param_value(out, &mut line_len, v.as_str())?;
+                ArchivedICalendarParameter::Role(v) => {
+                    write_param(out, &mut line_len, "ROLE", v)?;
                 }
-                ArchivedICalendarParameterValue::Duration(v) => {
-                    write!(out, "{v}")?;
+                ArchivedICalendarParameter::Rsvp(v) => {
+                    write_param(
+                        out,
+                        &mut line_len,
+                        "RSVP",
+                        if *v { "TRUE" } else { "FALSE" },
+                    )?;
+                }
+                ArchivedICalendarParameter::ScheduleAgent(v) => {
+                    write_param(out, &mut line_len, "SCHEDULE-AGENT", v)?;
+                }
+                ArchivedICalendarParameter::ScheduleForceSend(v) => {
+                    write_param(out, &mut line_len, "SCHEDULE-FORCE-SEND", v)?;
+                }
+                ArchivedICalendarParameter::ScheduleStatus(v) => {
+                    write_param(out, &mut line_len, "SCHEDULE-STATUS", v)?;
+                }
+                ArchivedICalendarParameter::SentBy(v) => {
+                    write_uri_param(out, &mut line_len, "SENT-BY", v)?;
+                }
+                ArchivedICalendarParameter::Tzid(v) => {
+                    write_param(out, &mut line_len, "TZID", v)?;
+                }
+                ArchivedICalendarParameter::Value(v) => {
+                    write_param(out, &mut line_len, "VALUE", v)?;
+                    value_type = Some(v);
+                }
+                ArchivedICalendarParameter::Display(v) => {
+                    write_params(out, &mut line_len, "DISPLAY", v)?;
+                }
+                ArchivedICalendarParameter::Email(v) => {
+                    write_param(out, &mut line_len, "EMAIL", v)?;
+                }
+                ArchivedICalendarParameter::Feature(v) => {
+                    write_params(out, &mut line_len, "FEATURE", v)?;
+                }
+                ArchivedICalendarParameter::Label(v) => {
+                    write_param(out, &mut line_len, "LABEL", v)?;
+                }
+                ArchivedICalendarParameter::Size(v) => {
+                    write!(out, "SIZE={}", v)?;
+                    line_len += 8;
+                }
+                ArchivedICalendarParameter::Filename(v) => {
+                    write_param(out, &mut line_len, "FILENAME", v)?;
+                }
+                ArchivedICalendarParameter::ManagedId(v) => {
+                    write_param(out, &mut line_len, "MANAGED-ID", v)?;
+                }
+                ArchivedICalendarParameter::Order(v) => {
+                    write!(out, "ORDER={}", v)?;
+                    line_len += 8;
+                }
+                ArchivedICalendarParameter::Schema(v) => {
+                    write_uri_param(out, &mut line_len, "SCHEMA", v)?;
+                }
+                ArchivedICalendarParameter::Derived(v) => {
+                    write_param(
+                        out,
+                        &mut line_len,
+                        "DERIVED",
+                        if *v { "TRUE" } else { "FALSE" },
+                    )?;
+                }
+                ArchivedICalendarParameter::Gap(v) => {
+                    write!(out, "GAP={}", v)?;
                     line_len += 14;
                 }
-                ArchivedICalendarParameterValue::Linkrel(v) => {
-                    write_param_value(out, &mut line_len, v.as_str())?;
+                ArchivedICalendarParameter::Linkrel(v) => {
+                    write_uri_param(out, &mut line_len, "LINKREL", v)?;
                 }
-                ArchivedICalendarParameterValue::Null => {
-                    last_param = None;
+                ArchivedICalendarParameter::Other(v) => {
+                    for (pos, item) in v.iter().enumerate() {
+                        if pos == 0 {
+                            write!(out, "{item}")?;
+                            line_len += item.len() + 1;
+                        } else {
+                            if pos == 1 {
+                                write!(out, "=")?;
+                            } else {
+                                write!(out, ",")?;
+                            }
+                            line_len += 1;
+
+                            write_param_value(out, &mut line_len, item)?;
+                        }
+                    }
                 }
             }
         }
@@ -197,7 +246,7 @@ impl ArchivedICalendarEntry {
 
                 let text = match value {
                     ArchivedICalendarValue::Binary(v) => {
-                        write_bytes(out, Some(&mut line_len), v)?;
+                        write_bytes(out, &mut line_len, v)?;
                         continue;
                     }
                     ArchivedICalendarValue::Boolean(v) => {
@@ -211,7 +260,7 @@ impl ArchivedICalendarEntry {
                         continue;
                     }
                     ArchivedICalendarValue::PartialDateTime(v) => {
-                        v.format_as_ical(out, types.unwrap_or(&default_type))?;
+                        v.format_as_ical(out, value_type.unwrap_or(&default_type))?;
                         line_len += 6;
                         continue;
                     }
@@ -241,7 +290,7 @@ impl ArchivedICalendarEntry {
                         continue;
                     }
                     ArchivedICalendarValue::Text(v) => {
-                        write_text(out, &mut line_len, v, true, true)?;
+                        write_value(out, &mut line_len, v)?;
                         continue;
                     }
                     ArchivedICalendarValue::CalendarScale(v) => v.as_str(),
@@ -264,6 +313,41 @@ impl ArchivedICalendarEntry {
     }
 }
 
+pub(crate) fn write_uri_param(
+    out: &mut impl Write,
+    line_len: &mut usize,
+    name: &str,
+    value: &ArchivedUri,
+) -> std::fmt::Result {
+    write!(out, "{}=\"", name)?;
+    *line_len += name.len() + 3;
+    write_uri(out, line_len, value, false)?;
+    write!(out, "\"")
+}
+
+pub(crate) fn write_uri_params(
+    out: &mut impl Write,
+    line_len: &mut usize,
+    name: &str,
+    values: &[ArchivedUri],
+) -> std::fmt::Result {
+    write!(out, "{}", name)?;
+    *line_len += name.len() + 1;
+
+    for (pos, v) in values.iter().enumerate() {
+        if pos > 0 {
+            write!(out, ",\"")?;
+        } else {
+            write!(out, "=\"")?;
+        }
+        *line_len += 3;
+        write_uri(out, line_len, v, false)?;
+        write!(out, "\"")?;
+    }
+
+    Ok(())
+}
+
 pub(crate) fn write_uri(
     out: &mut impl Write,
     line_len: &mut usize,
@@ -284,9 +368,9 @@ pub(crate) fn write_uri(
                 write!(out, "base64,")?;
             }
             *line_len += 8;
-            write_bytes(out, Some(line_len), &v.data)
+            write_bytes(out, line_len, &v.data)
         }
-        ArchivedUri::Location(v) => write_text(out, line_len, v, true, true),
+        ArchivedUri::Location(v) => write_value(out, line_len, v),
     }
 }
 
@@ -295,14 +379,7 @@ impl Display for ArchivedICalendarRecurrenceRule {
         write!(f, "FREQ={}", self.freq.as_str())?;
         if let Some(until) = self.until.as_ref() {
             write!(f, ";UNTIL=")?;
-            until.format_as_ical(
-                f,
-                if until.has_date_and_time() {
-                    &ArchivedICalendarValueType::DateTime
-                } else {
-                    &ArchivedICalendarValueType::Date
-                },
-            )?;
+            until.format_as_ical(f, &ArchivedICalendarValueType::DateTime)?;
         }
         if let Some(count) = self.count.as_ref().filter(|c| **c > 0) {
             write!(f, ";COUNT={}", count)?;
@@ -394,12 +471,6 @@ impl Display for ArchivedICalendarRecurrenceRule {
         if let Some(wkst) = self.wkst.as_ref() {
             write!(f, ";WKST={}", wkst.as_str())?;
         }
-        if let Some(rscale) = self.rscale.as_ref() {
-            write!(f, ";RSCALE={}", rscale.as_str())?;
-        }
-        if let Some(skip) = self.skip.as_ref() {
-            write!(f, ";SKIP={}", skip.as_str())?;
-        }
 
         Ok(())
     }
@@ -414,16 +485,6 @@ impl Display for ArchivedICalendarDay {
     }
 }
 
-impl Display for ArchivedICalendarMonth {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        if !self.is_leap() {
-            write!(f, "{}", self.month())
-        } else {
-            write!(f, "{}L", self.month())
-        }
-    }
-}
-
 impl Display for ArchivedICalendarPeriod {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         match self {
